PAD, BEGIN, END, UNKNOWN = 0, 1, 2, 3

AUXILIARY = [
    'PAD',
    'BEGIN',
    'END',
    'UNKNOWN'
]

TARGET_SYMBOLS = [
    "PAD",
    "BEGIN",
    "END",
    "UNKNOWN",
    "-",
    "B-END",
    "B-POSTFIX",
    "B-PREF",
    "B-ROOT",
    "B-SUFF",
    "E-END",
    "E-POSTFIX",
    "E-PREF",
    "E-ROOT",
    "E-SUFF",
    "M-END",
    "M-POSTFIX",
    "M-PREF",
    "M-ROOT",
    "M-SUFF",
    "S-END",
    "S-HYPH",
    "S-LINK",
    "S-PREF",
    "S-ROOT",
    "S-SUFF"
]
